from typing import List

import pandas as pd

from vivarium_inputs.globals import gbd, SEXES, NON_MAX_TMREL
from gbd_mapping import RiskFactor


def get_estimation_years(*_, **__) -> pd.Series:
    data = gbd.get_estimation_years()
    return data


def get_year_block(*_, **__) -> pd.DataFrame:
    estimation_years = get_estimation_years()
    year_block = pd.DataFrame({'year_start': range(min(estimation_years), max(estimation_years) + 1)})
    year_block['year_end'] = year_block['year_start'] + 1
    return year_block


def get_age_group_ids(*_, **__) -> List[int]:
    data = gbd.get_age_group_id()
    return data


def get_age_bins(*_, **__) -> pd.DataFrame:
    age_bins = (
        gbd.get_age_bins()[['age_group_id', 'age_group_name', 'age_group_years_start', 'age_group_years_end']]
            .rename(columns={'age_group_years_start': 'age_start',
                             'age_group_years_end': 'age_end'})
    )
    return age_bins


def get_location_id(location_name):
    return {r.location_name: r.location_id for _, r in gbd.get_location_ids().iterrows()}[location_name]


def get_location_id_parents(location_id: int) -> List[int]:
    location_metadata = gbd.get_location_path_to_global().set_index('location_id')
    parent_ids = [int(loc) for loc in location_metadata.at[location_id, 'path_to_top_parent'].split(',')]
    return parent_ids


def get_demographic_dimensions(location_id: int, draws: bool = False, value: float = None) -> pd.DataFrame:
    ages = get_age_group_ids()
    estimation_years = get_estimation_years()
    years = range(min(estimation_years), max(estimation_years) + 1)
    sexes = [SEXES['Male'], SEXES['Female']]
    location = [location_id]
    values = [location, sexes, ages, years]
    names = ['location_id', 'sex_id', 'age_group_id', 'year_id']

    data = (pd.MultiIndex
            .from_product(values, names=names)
            .to_frame(index=False))
    if draws:
        for i in range(1000):
            data[f'draw_{i}'] = value
    return data


def get_tmrel_category(entity: RiskFactor) -> str:
    if entity.name in NON_MAX_TMREL:
        tmrel_cat = NON_MAX_TMREL[entity.name]
    else:
        tmrel_cat = sorted(list(entity.categories.to_dict()), key=lambda x: int(x[3:]))[-1]
    return tmrel_cat



import gbd_mapping
import numpy as np
import pandas as pd
import pytest
from enum import IntFlag

from gbd_mapping import causes, risk_factors, covariates, ModelableEntity
from vivarium_inputs import extract, utility_data
from tests.extract.check import RUNNING_ON_CI


pytestmark = pytest.mark.skipif(RUNNING_ON_CI, reason="Don't run these tests on the CI server")


VALIDATE_FLAG = False


def success_expected(entity_name, measure_name, location):
    df = extract.extract_data(entity_name, measure_name, location, validate=VALIDATE_FLAG)
    return df


def fail_expected(entity_name, measure_name, location):
    with pytest.raises(Exception):
        df = extract.extract_data(entity_name, measure_name, location, validate=VALIDATE_FLAG)


class MCFlag(IntFlag):
    """
        Use the idea of a bit field with support from python's enum type IntFlag
        See here for general information on bit fields:
        https://en.wikipedia.org/wiki/Bit_field

        And here for python's enum type:
        https://docs.python.org/3.6/library/enum.html
    """
    INCIDENCE_RATE = 1
    PREVALENCE = 2
    BIRTH_PREVALENCE = 4
    DISABILITY_WEIGHT = 8
    REMISSION_RATE = 16
    DEATHS = 32
    ALL = INCIDENCE_RATE | PREVALENCE | BIRTH_PREVALENCE | DISABILITY_WEIGHT | REMISSION_RATE | DEATHS


entity = [
    (causes.measles, MCFlag.INCIDENCE_RATE
        | MCFlag.PREVALENCE
        | MCFlag.BIRTH_PREVALENCE
        | MCFlag.REMISSION_RATE
        | MCFlag.DEATHS),
    (causes.diabetes_mellitus_type_2,
        MCFlag.INCIDENCE_RATE
        | MCFlag.PREVALENCE
        | MCFlag.BIRTH_PREVALENCE
        | MCFlag.DEATHS),
    ]
measures = [
    ('incidence_rate', MCFlag.INCIDENCE_RATE),
    ('prevalence', MCFlag.PREVALENCE),
    ('birth_prevalence', MCFlag.BIRTH_PREVALENCE),
    ('disability_weight', MCFlag.DISABILITY_WEIGHT),
    ('remission_rate', MCFlag.REMISSION_RATE),
    ('deaths', MCFlag.DEATHS)]
locations = ['India']

@pytest.mark.parametrize('entity', entity, ids=lambda x: x[0].name)
@pytest.mark.parametrize('measure', measures, ids=lambda x: x[0])
@pytest.mark.parametrize('location', locations)
def test_extract_causelike(entity, measure, location):
    entity_name, entity_expected_measure_ids = entity
    measure_name, measure_id = measure
    tester = success_expected if (entity_expected_measure_ids & measure_id) else fail_expected
    df = tester(entity_name, measure_name, utility_data.get_location_id(location))


class MRFlag(IntFlag):
    """
        Use the idea of a bit field with support from python's enum type IntFlag
        See here for general information on bit fields:
        https://en.wikipedia.org/wiki/Bit_field

        And here for python's enum type:
        https://docs.python.org/3.6/library/enum.html
    """
    EXPOSURE = 1
    EXPOSURE_SD = 2
    EXPOSURE_DIST_WEIGHTS = 4
    RELATIVE_RISK = 8
    PAF = 16
    ETIOLOGY_PAF = 32
    MEDIATION_FACTORS = 64
    ALL = (EXPOSURE | EXPOSURE_SD | EXPOSURE_DIST_WEIGHTS
            | RELATIVE_RISK | PAF | ETIOLOGY_PAF | MEDIATION_FACTORS)


entity_r = [
    (risk_factors.high_systolic_blood_pressure,
        MRFlag.EXPOSURE | MRFlag.EXPOSURE_SD | MRFlag.EXPOSURE_DIST_WEIGHTS
        | MRFlag.RELATIVE_RISK | MRFlag.PAF | MRFlag.ETIOLOGY_PAF | MRFlag.MEDIATION_FACTORS),
    (risk_factors.low_birth_weight_and_short_gestation,
        MRFlag.EXPOSURE | MRFlag.RELATIVE_RISK | MRFlag.PAF | MRFlag.ETIOLOGY_PAF)]
measures_r = [
    ('exposure', MRFlag.EXPOSURE),
    ('exposure_standard_deviation', MRFlag.EXPOSURE_SD),
    ('exposure_distribution_weights', MRFlag.EXPOSURE_DIST_WEIGHTS),
    ('relative_risk', MRFlag.RELATIVE_RISK),
    ('population_attributable_fraction', MRFlag.PAF),
    ('etiology_population_attributable_fraction', MRFlag.ETIOLOGY_PAF),
    ('mediation_factors', MRFlag.MEDIATION_FACTORS)]
locations_r = ['India']
@pytest.mark.parametrize('entity', entity_r, ids=lambda x: x[0].name)
@pytest.mark.parametrize('measure', measures_r, ids=lambda x: x[0])
@pytest.mark.parametrize('location', locations_r)
def test_extract_risklike(entity, measure, location):
    entity_name, entity_expected_measure_ids = entity
    measure_name, measure_id = measure
    tester = success_expected if (entity_expected_measure_ids & measure_id) else fail_expected
    df = tester(entity_name, measure_name, utility_data.get_location_id(location))


entity_cov = [
    covariates.systolic_blood_pressure_mmhg,
]   
measures_cov = [
    'estimate'
]
locations_cov = ['India']
@pytest.mark.parametrize('entity', entity_cov)
@pytest.mark.parametrize('measure', measures_cov)
@pytest.mark.parametrize('location', locations_cov)
def test_extract_covariatelike(entity, measure, location):
    df = extract.extract_data(entity, measure, utility_data.get_location_id(location), validate=VALIDATE_FLAG)


@pytest.mark.parametrize('measures',
    ['structure',
    'theoretical_minimum_risk_life_expectancy'])
def test_extract_population(measures):
    pop = ModelableEntity('ignored', 'population', None)
    df = extract.extract_data(pop, measures, utility_data.get_location_id('India'), validate=VALIDATE_FLAG)



from vivarium_inputs.interface import (get_measure, get_population_structure, get_theoretical_minimum_risk_life_expectancy,
                                       get_age_bins, get_demographic_dimensions, get_raw_data)

from vivarium_inputs.__about__ import (__author__, __copyright__, __email__, __license__,
                                       __summary__, __title__, __uri__, __version__, )



import pytest

import pandas as pd
import numpy as np

from vivarium_inputs.validation import sim
from vivarium_inputs.globals import DataTransformationError


@pytest.fixture
def mock_validation_context():
    years = pd.DataFrame({'year_start': range(1990, 2017),
                          'year_end': range(1991, 2018)})
    age_bins = pd.DataFrame({'age_group_id': [1, 2, 3, 4, 5],
                             'age_group_name': ['youngest', 'young', 'middle', 'older', 'oldest'],
                             'age_start': [0, 1, 15, 45, 60],
                             'age_end': [1, 15, 45, 60, 100]})
    context = sim.SimulationValidationContext(
        location='United States',
        years=years,
        age_bins=age_bins
    )

    return context


@pytest.mark.parametrize("location", ("Kenya", "Papua New Guinea"))
def test__validate_location_column_pass(mock_validation_context, location):
    mock_validation_context['location'] = location
    df = pd.DataFrame({'location': [location], 'other': [0]}).set_index(['location', 'other'])
    sim.validate_location_column(df, mock_validation_context)


@ pytest.mark.parametrize('locations,expected_location', (
        (['Kenya', 'Kenya'], 'Egypt'),
        (['Algeria', 'Nigeria'], 'Algeria')
), ids=('mismatch', 'multiple'))
def test_validate_location_column_fail(mock_validation_context, locations, expected_location):
    mock_validation_context['location'] = expected_location
    df = pd.DataFrame({'location': locations, 'other': [0]*len(locations)}).set_index(['location', 'other'])
    with pytest.raises(DataTransformationError):
        sim.validate_location_column(df, mock_validation_context)


def test_validate_sex_column_pass():
    df = pd.DataFrame({'sex': ['Male', 'Female'], 'other': [0]*2}).set_index(['sex', 'other'])
    sim.validate_sex_column(df)


@pytest.mark.parametrize("sexes", (
        ('Male', 'female'),
        ('Male', 'Male'),
        ('Female', 'Female')
), ids=('lowercase', 'missing_female', 'missing_male'))
def test_validate_sex_column_fail(sexes):
    df = pd.DataFrame({'sex': sexes, 'other': [0]*len(sexes)}).set_index(['sex', 'other'])
    with pytest.raises(DataTransformationError):
        sim.validate_sex_column(df)


def test_validate_age_columns_pass(mock_validation_context):
    ages = (mock_validation_context['age_bins']
            .filter(['age_start', 'age_end']))
    # Shuffle the rows and set index
    ages = ages.sample(frac=1).reset_index(drop=True)
    ages = ages.set_index(pd.IntervalIndex.from_arrays(ages.age_start, ages.age_end,
                                                       closed='left', name='age'),
                          append=True)
    sim.validate_age_column(ages, mock_validation_context)


def test_validate_age_columns_invalid_age(mock_validation_context):
    ages = (mock_validation_context['age_bins']
            .filter(['age_start', 'age_end']).copy())
    ages.loc[2, 'age_start'] = -1
    ages = ages.set_index(pd.IntervalIndex.from_arrays(ages.age_start, ages.age_end,
                                                       closed='left', name='age'),
                          append=True)
    with pytest.raises(DataTransformationError):
        sim.validate_age_column(ages, mock_validation_context)


def test_validate_age_columns_missing_group(mock_validation_context):
    ages = (mock_validation_context['age_bins']
            .filter(['age_start', 'age_end']))
    ages = ages.drop(2)
    ages = ages.set_index(pd.IntervalIndex.from_arrays(ages.age_start, ages.age_end,
                                                       closed='left', name='age'),
                          append=True)
    with pytest.raises(DataTransformationError):
        sim.validate_age_column(ages, mock_validation_context)


def test_validate_year_columns_pass(mock_validation_context):
    years = mock_validation_context['years']
    # Shuffle the rows and set index
    years = years.sample(frac=1).reset_index(drop=True)
    years = years.set_index(pd.IntervalIndex.from_arrays(years.year_start, years.year_end,
                                                         closed='left', name='year'),
                            append=True)
    sim.validate_year_column(years, mock_validation_context)


def test_validate_year_columns_invalid_year(mock_validation_context):
    years = mock_validation_context['years'].copy()
    years.loc[2, 'year_end'] = 20100
    # Shuffle the rows and set index
    years = years.sample(frac=1).reset_index(drop=True)
    years = years.set_index(pd.IntervalIndex.from_arrays(years.year_start, years.year_end,
                                                         closed='left', name='year'),
                            append=True)
    with pytest.raises(DataTransformationError):
        sim.validate_year_column(years, mock_validation_context)


def test_validate_year_columns_missing_group(mock_validation_context):
    years = mock_validation_context['years'].sort_values(['year_start', 'year_end'])
    years = years.drop(0)
    years = years.set_index(pd.IntervalIndex.from_arrays(years.year_start, years.year_end,
                                                         closed='left', name='year'),
                            append=True)
    with pytest.raises(DataTransformationError):
        sim.validate_year_column(years, mock_validation_context)


@pytest.mark.parametrize("values", [(-1, 2, 3)], ids=['integers'])
def test_validate_value_column_pass(values):
    df = pd.DataFrame({'value': values})
    sim.validate_value_column(df)


@pytest.mark.parametrize("values", [(1, 2, np.inf),
                                    (1, np.nan, 2)],
                         ids=["infinity", "missing"])
def test_validate_value_column_fail(values):
    df = pd.DataFrame({'value': values})
    with pytest.raises(DataTransformationError):
        sim.validate_value_column(df)


@pytest.mark.parametrize('values, ids, restriction_type, fill', [
        ((1, 1, 1, 1, 1), (1, 5), 'outer', 0.0),
        ((0, 0, 1, 1, 1), (3, 5), 'outer', 0.0),
        ((0, 1, 1, 1, 0), (2, 4), 'outer', 0.0),
        ((1, 1, 1, 0, 0), (1, 3), 'outer', 0.0),
        ((2, 2, 2, 1, 1), (1, 3), 'outer', 1.0),
], ids=('no_restr', 'left_restr', 'outer_restr', 'right_restr', 'nonzero_fill'))
def test_check_age_restrictions(mocker, mock_validation_context, values, ids, restriction_type, fill):
    entity = mocker.patch('vivarium_inputs.validation.sim.utilities.get_age_group_ids_by_restriction')
    entity.return_value = ids
    age_bins = mock_validation_context['age_bins']
    idx = pd.IntervalIndex.from_arrays(age_bins.age_start, age_bins.age_end, closed='left', name='age')
    df = pd.DataFrame({'value': values}, index=idx)
    sim.check_age_restrictions(df, entity, restriction_type, fill, mock_validation_context)


@pytest.mark.parametrize('values, ids, restriction_type, fill', [
        ((1, 1, 1, 1, 1), (1, 4), 'outer', 0.0),
        ((0, 1, 1, 1, 1), (1, 3), 'outer', 0.0),
        ((1, 1, 1, 1, 0), (1, 3), 'outer', 0.0),
        ((2, 2, 2, 2, 1), (2, 5), 'outer', 1.0),
], ids=('both_sides', 'left_side', 'right_side', 'nonzero_fill'))
def test_check_age_restrictions_fail(mocker, mock_validation_context, values, ids, restriction_type, fill):
    entity = mocker.patch('vivarium_inputs.validation.sim.utilities.get_age_group_ids_by_restriction')
    entity.return_value = ids
    age_bins = mock_validation_context['age_bins']
    idx = pd.IntervalIndex.from_arrays(age_bins.age_start, age_bins.age_end, closed='left', name='age')
    df = pd.DataFrame({'value': values}, index=idx)
    with pytest.raises(DataTransformationError):
        sim.check_age_restrictions(df, entity, restriction_type, fill, mock_validation_context)


@pytest.mark.parametrize('values, restrictions, fill', [
    ((1, 1, 1, 1), (False, False), 0.0),
    ((1, 1, 0, 0), (True, False), 0.0),
    ((0, 0, 1, 1), (False, True), 0.0),
    ((1, 1, 2, 2), (False, True), 1.0)
], ids=('None', 'male', 'female', 'nonzero_fill'))
def test_check_sex_restrictions(values, restrictions, fill):
    df = pd.DataFrame({'sex': ['Male', 'Male', 'Female', 'Female'], 'other': [0]*4, 'value': values})
    df = df.set_index(['sex', 'other'])
    sim.check_sex_restrictions(df, restrictions[0], restrictions[1], fill)


@pytest.mark.parametrize('values, restrictions, fill', [
    ((1, 1, 1, 0), (True, False), 0.0),
    ((0, 1, 1, 1), (False, True), 0.0),
    ((1, 2, 2, 2), (False, True), 1.0)
], ids=('male', 'female', 'nonzero_fill'))
def test_check_sex_restrictions_fail(values, restrictions, fill):
    df = pd.DataFrame({'sex': ['Male', 'Male', 'Female', 'Female'], 'other': [0]*4, 'value': values})
    df = df.set_index(['sex', 'other'])
    with pytest.raises(DataTransformationError):
        sim.check_sex_restrictions(df, restrictions[0], restrictions[1], fill)


@pytest.mark.parametrize('df', [pd.DataFrame(columns=['a', 'b', 'c', 'd']),
                                pd.DataFrame({'a': [0], 'b': [0], 'c': [0], 'd': [0]}),
                                pd.DataFrame({'a': [1]*3, 'b': [2]*3, 'c': [3]*3, 'd': [np.NaN]*3})])
def test_validate_expected_index_columns_pass(df):
    cols = ['a', 'b', 'c', 'd']
    sim.validate_expected_index_and_columns([None], df.index.names, cols, df.columns)
    df = df.set_index(cols[:2])
    sim.validate_expected_index_and_columns(cols[:2], df.index.names, cols[2:], df.columns)
    df = df.set_index(cols[2:], append=True)
    sim.validate_expected_index_and_columns(cols, df.index.names, [], df.columns)


@pytest.mark.parametrize('df', [pd.DataFrame(columns=['a', 'b', 'c', 'd']),
                                pd.DataFrame({'a': [0], 'b': [0], 'c': [0], 'd': [0]}),
                                pd.DataFrame({'a': [1]*3, 'b': [2]*3, 'c': [3]*3, 'd': [np.NaN]*3})])
def test_validate_expected_columns_fail(df):
    cols = ['a', 'b', 'c', 'd']
    with pytest.raises(DataTransformationError, match='missing columns'):
        sim.validate_expected_index_and_columns([None], df.index.names, cols + ['e'], df.columns)
    with pytest.raises(DataTransformationError, match='missing index names'):
        data = df.set_index(cols[:2])
        sim.validate_expected_index_and_columns(cols[:3], data.index.names, cols[:2], data.columns)

    with pytest.raises(DataTransformationError, match='extra columns'):
        sim.validate_expected_index_and_columns([None], df.index.names, cols[1:], df.columns)
    with pytest.raises(DataTransformationError, match='extra index names'):
        data = df.set_index(cols[:2])
        sim.validate_expected_index_and_columns(cols[:1], data.index.names, cols[:2], data.columns)


import pytest
import pandas as pd

import gbd_mapping as gm

from vivarium_inputs import utility_data as ud
from tests.extract.check import RUNNING_ON_CI


pytestmark = pytest.mark.skipif(RUNNING_ON_CI, reason="Don't run these tests on the CI server")


def test_get_estimation_years():
    result = ud.get_estimation_years()
    assert(set([1990, 1995, 2000, 2005, 2010, 2015, 2017, 2019]) == set(result))


def test_get_year_block():
    result = ud.get_year_block()
    assert(result.year_start.iloc[0] == 1990)
    assert(result.year_end.iloc[-1] == 2020)
    

def test_get_age_group_ids():
    result = ud.get_age_group_ids()
    truth = list(range(2, 21)) + [30, 31, 32, 235]
    assert(result == truth)


def test_get_location_id():
    result_india = ud.get_location_id('India')
    assert(163 == result_india)


def test_get_location_id_parents():
    assert([1, 158, 159, 163] == ud.get_location_id_parents(163))


def test_get_demographic_dimensions():
    result = ud.get_demographic_dimensions(163)
    assert(len(result))
    assert(set(['location_id', 'sex_id', 'age_group_id', 'year_id']) == set(result.columns))


def test_get_tmrel_category():
    result = ud.get_tmrel_category(gm.risk_factors.smoking)
    assert('cat2' == result)


from typing import Union
from itertools import product

from gbd_mapping import Cause, Sequela, RiskFactor, Etiology, Covariate, causes
import pandas as pd
import numpy as np
from loguru import logger

from vivarium_inputs import utilities, extract, utility_data
from vivarium_inputs.globals import (InvalidQueryError, DEMOGRAPHIC_COLUMNS, MEASURES, SEXES, DRAW_COLUMNS,
                                     Population, DataDoesNotExistError, EXTRA_RESIDUAL_CATEGORY)
from vivarium_inputs.mapping_extension import AlternativeRiskFactor, HealthcareEntity, HealthTechnology


DISTRIBUTION_COLUMNS = ['exp', 'gamma', 'invgamma', 'llogis', 'gumbel', 'invweibull', 'weibull',
                         'lnorm', 'norm', 'glnorm', 'betasr', 'mgamma', 'mgumbel']
COVARIATE_VALUE_COLUMNS = ['mean_value', 'upper_value', 'lower_value']


def get_data(entity, measure: str, location: Union[str, int]):
    measure_handlers = {
        # Cause-like measures
        'incidence_rate': (get_incidence_rate, ('cause', 'sequela')),
        'raw_incidence_rate': (get_raw_incidence_rate, ('cause', 'sequela')),
        'prevalence': (get_prevalence, ('cause', 'sequela')),
        'birth_prevalence': (get_birth_prevalence, ('cause', 'sequela')),
        'disability_weight': (get_disability_weight, ('cause', 'sequela')),
        'remission_rate': (get_remission_rate, ('cause',)),
        'cause_specific_mortality_rate': (get_cause_specific_mortality_rate, ('cause',)),
        'excess_mortality_rate': (get_excess_mortality_rate, ('cause',)),
        'deaths': (get_deaths, ('cause',)),
        # Risk-like measures
        'exposure': (get_exposure, ('risk_factor', 'alternative_risk_factor',)),
        'exposure_standard_deviation': (get_exposure_standard_deviation, ('risk_factor', 'alternative_risk_factor')),
        'exposure_distribution_weights': (get_exposure_distribution_weights, ('risk_factor', 'alternative_risk_factor')),
        'relative_risk': (get_relative_risk, ('risk_factor',)),
        'population_attributable_fraction': (get_population_attributable_fraction, ('risk_factor', 'etiology')),
        # Covariate measures
        'estimate': (get_estimate, ('covariate',)),
        # Population measures
        'structure': (get_structure, ('population',)),
        'theoretical_minimum_risk_life_expectancy': (get_theoretical_minimum_risk_life_expectancy, ('population',)),
        'age_bins': (get_age_bins, ('population',)),
        'demographic_dimensions': (get_demographic_dimensions, ('population',))

    }

    if measure not in measure_handlers:
        raise InvalidQueryError(f'No functions available to pull data for measure {measure}.')

    handler, entity_types = measure_handlers[measure]

    if entity.kind not in entity_types:
        raise InvalidQueryError(f'{measure.capitalize()} not available for {entity.kind}.')

    location_id = utility_data.get_location_id(location) if isinstance(location, str) else location
    data = handler(entity, location_id)

    if measure in ['structure', 'theoretical_minimum_risk_life_expectancy',
                   'estimate', 'exposure_distribution_weights']:
        value_cols = ['value']
    else:
        value_cols = DRAW_COLUMNS

    data = utilities.reshape(data, value_cols=value_cols)

    return data


def get_raw_incidence_rate(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'incidence_rate', location_id)
    if entity.kind == 'cause':
        restrictions_entity = entity
    else:  # sequela
        cause = [c for c in causes if c.sequelae and entity in c.sequelae][0]
        restrictions_entity = cause

    data = utilities.filter_data_by_restrictions(data, restrictions_entity,
                                                 'yld', utility_data.get_age_group_ids())
    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_incidence_rate(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = get_data(entity, 'raw_incidence_rate', location_id)
    prevalence = get_data(entity, 'prevalence', location_id)
    # Convert from "True incidence" to the incidence rate among susceptibles
    data /= 1 - prevalence
    return data.fillna(0)


def get_prevalence(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'prevalence', location_id)
    if entity.kind == 'cause':
        restrictions_entity = entity
    else:  # sequela
        cause = [c for c in causes if c.sequelae and entity in c.sequelae][0]
        restrictions_entity = cause

    data = utilities.filter_data_by_restrictions(data, restrictions_entity,
                                                 'yld', utility_data.get_age_group_ids())
    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_birth_prevalence(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'birth_prevalence', location_id)
    data = data.filter(['year_id', 'sex_id', 'location_id'] + DRAW_COLUMNS)
    data = utilities.normalize(data, fill_value=0)
    return data


def get_disability_weight(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    if entity.kind == 'cause':
        data = utility_data.get_demographic_dimensions(location_id, draws=True, value=0.0)
        data = data.set_index(utilities.get_ordered_index_cols(data.columns.difference(DRAW_COLUMNS)))
        if entity.sequelae:
            for sequela in entity.sequelae:
                try:
                    prevalence = get_data(sequela, 'prevalence', location_id)
                except DataDoesNotExistError:
                    # sequela prevalence does not exist so no point continuing with this sequela
                    continue
                disability = get_data(sequela, 'disability_weight', location_id)
                disability.index = disability.index.set_levels([location_id], 'location_id')
                data += prevalence * disability
        cause_prevalence = get_data(entity, 'prevalence', location_id)
        data = (data / cause_prevalence).fillna(0).reset_index()
    else:  # entity.kind == 'sequela'  
        try:
            data = extract.extract_data(entity, 'disability_weight', location_id)
            data = utilities.normalize(data)
            cause = [c for c in causes if c.sequelae and entity in c.sequelae][0]
            data = utilities.clear_disability_weight_outside_restrictions(data, cause, 0.0,
                                                                            utility_data.get_age_group_ids())
            data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
        except (IndexError, DataDoesNotExistError):
            logger.warning(f"{entity.name.capitalize()} has no disability weight data. All values will be 0.")
            data = utility_data.get_demographic_dimensions(location_id, draws=True, value=0.0)
    return data


def get_remission_rate(entity: Cause, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'remission_rate', location_id)

    data = utilities.filter_data_by_restrictions(data, entity,
                                                 'yld', utility_data.get_age_group_ids())
    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_cause_specific_mortality_rate(entity: Cause, location_id: int) -> pd.DataFrame:
    deaths = get_data(entity, 'deaths', location_id)  # population isn't by draws
    pop = get_data(Population(), 'structure', location_id)
    data = deaths.join(pop, lsuffix='_deaths', rsuffix='_pop')
    data[DRAW_COLUMNS] = data[DRAW_COLUMNS].divide(data.value, axis=0)
    return data.drop(['value'], 'columns')


def get_excess_mortality_rate(entity: Cause, location_id: int) -> pd.DataFrame:
    csmr = get_data(entity, 'cause_specific_mortality_rate', location_id)
    prevalence = get_data(entity, 'prevalence', location_id)
    data = (csmr / prevalence).fillna(0)
    data = data.replace([np.inf, -np.inf], 0)
    return data


def get_deaths(entity: Cause, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'deaths', location_id)
    data = utilities.filter_data_by_restrictions(data, entity,
                                                 'yll', utility_data.get_age_group_ids())
    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_exposure(entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'exposure', location_id)
    data = data.drop('modelable_entity_id', 'columns')

    if entity.name in EXTRA_RESIDUAL_CATEGORY:
        data = data[data.parameter != EXTRA_RESIDUAL_CATEGORY[entity.name]]

    if entity.kind in ['risk_factor', 'alternative_risk_factor']:
        data = utilities.filter_data_by_restrictions(data, entity,
                                                     'outer', utility_data.get_age_group_ids())

    if entity.distribution in ['dichotomous', 'ordered_polytomous', 'unordered_polytomous']:
        tmrel_cat = utility_data.get_tmrel_category(entity)
        exposed = data[data.parameter != tmrel_cat]
        unexposed = data[data.parameter == tmrel_cat]

        #  FIXME: We fill 1 as exposure of tmrel category, which is not correct.
        data = pd.concat([utilities.normalize(exposed, fill_value=0), utilities.normalize(unexposed, fill_value=1)],
                         ignore_index=True)

        # normalize so all categories sum to 1
        cols = list(set(data.columns).difference(DRAW_COLUMNS + ['parameter']))
        sums = data.groupby(cols)[DRAW_COLUMNS].sum()
        data = (data.groupby('parameter')
                .apply(lambda df: df.set_index(cols).loc[:, DRAW_COLUMNS].divide(sums))
                .reset_index())
    else:
        data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS + ['parameter'])
    return data


def get_exposure_standard_deviation(entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'exposure_standard_deviation', location_id)
    data = data.drop('modelable_entity_id', 'columns')

    exposure = extract.extract_data(entity, 'exposure', location_id)
    valid_age_groups = utilities.get_exposure_and_restriction_ages(exposure, entity)
    data = data[data.age_group_id.isin(valid_age_groups)]

    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_exposure_distribution_weights(entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'exposure_distribution_weights', location_id)

    exposure = extract.extract_data(entity, 'exposure', location_id)
    valid_ages = utilities.get_exposure_and_restriction_ages(exposure, entity)

    data.drop('age_group_id', axis=1, inplace=True)
    df = []
    for age_id in valid_ages:
        copied = data.copy()
        copied['age_group_id'] = age_id
        df.append(copied)
    data = pd.concat(df)
    data = utilities.normalize(data, fill_value=0, cols_to_fill=DISTRIBUTION_COLUMNS)
    data = data.filter(['location_id', 'sex_id', 'age_group_id', 'year_id'] + DISTRIBUTION_COLUMNS)
    data = utilities.wide_to_long(data, DISTRIBUTION_COLUMNS, var_name='parameter')
    return data


def filter_relative_risk_to_cause_restrictions(data: pd.DataFrame) -> pd.DataFrame:
    """ It applies age restrictions according to affected causes
    and affected measures. If affected measure is incidence_rate,
    it applies the yld_age_restrictions. If affected measure is
    excess_mortality_rate, it applies the yll_age_restrictions to filter
    the relative_risk data"""

    causes_map = {c.name: c for c in causes}
    temp = []
    affected_entities = set(data.affected_entity)
    affected_measures = set(data.affected_measure)
    for cause, measure in product(affected_entities, affected_measures):
        df = data[(data.affected_entity == cause) & (data.affected_measure == measure)]
        cause = causes_map[cause]
        if measure == 'excess_mortality_rate':
            start, end = utilities.get_age_group_ids_by_restriction(cause, 'yll')
        else:  # incidence_rate
            start, end = utilities.get_age_group_ids_by_restriction(cause, 'yld')
        temp.append(df[df.age_group_id.isin(range(start, end + 1))])
    data = pd.concat(temp)
    return data


def get_relative_risk(entity: RiskFactor, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'relative_risk', location_id)

    # FIXME: we don't currently support yll-only causes so I'm dropping them because the data in some cases is
    #  very messed up, with mort = morb = 1 (e.g., aortic aneurysm in the RR data for high systolic bp) -
    #  2/8/19 K.W.
    yll_only_causes = set([c.gbd_id for c in causes if c.restrictions.yll_only])
    data = data[~data.cause_id.isin(yll_only_causes)]

    data = utilities.convert_affected_entity(data, 'cause_id')
    morbidity = data.morbidity == 1
    mortality = data.mortality == 1
    data.loc[morbidity & mortality, 'affected_measure'] = 'incidence_rate'
    data.loc[morbidity & ~mortality, 'affected_measure'] = 'incidence_rate'
    data.loc[~morbidity & mortality, 'affected_measure'] = 'excess_mortality_rate'
    data = filter_relative_risk_to_cause_restrictions(data)

    data = data.filter(DEMOGRAPHIC_COLUMNS + ['affected_entity', 'affected_measure', 'parameter'] + DRAW_COLUMNS)
    data = data.groupby(['affected_entity', 'parameter']).apply(utilities.normalize, fill_value=1).reset_index(drop=True)

    if entity.distribution in ['dichotomous', 'ordered_polytomous', 'unordered_polytomous']:
        tmrel_cat = utility_data.get_tmrel_category(entity)
        tmrel_mask = data.parameter == tmrel_cat
        data.loc[tmrel_mask, DRAW_COLUMNS] = (data.loc[tmrel_mask, DRAW_COLUMNS]
                                              .mask(np.isclose(data.loc[tmrel_mask, DRAW_COLUMNS], 1.0), 1.0))

    return data


def filter_by_relative_risk(df: pd.DataFrame, relative_risk: pd.DataFrame) -> pd.DataFrame:
    c_id = df.cause_id.unique()[0]
    rr = relative_risk[relative_risk.cause_id == c_id]
    #  We presume all attributable mortality moves through incidence.
    if set(rr.mortality) == {1} and set(rr.morbidity) == {1}:
        df = df[df.measure_id == MEASURES['YLDs']]
    return df


def get_population_attributable_fraction(entity: Union[RiskFactor, Etiology], location_id: int) -> pd.DataFrame:
    causes_map = {c.gbd_id: c for c in causes}
    if entity.kind == 'risk_factor':
        data = extract.extract_data(entity, 'population_attributable_fraction', location_id)
        relative_risk = extract.extract_data(entity, 'relative_risk', location_id)

        # FIXME: we don't currently support yll-only causes so I'm dropping them because the data in some cases is
        #  very messed up, with mort = morb = 1 (e.g., aortic aneurysm in the RR data for high systolic bp) -
        #  2/8/19 K.W.
        yll_only_causes = set([c.gbd_id for c in causes if c.restrictions.yll_only])
        data = data[~data.cause_id.isin(yll_only_causes)]
        relative_risk = relative_risk[~relative_risk.cause_id.isin(yll_only_causes)]

        data = (data.groupby('cause_id', as_index=False)
                .apply(filter_by_relative_risk, relative_risk)
                .reset_index(drop=True))

        temp = []
        # We filter paf age groups by cause level restrictions.
        for (c_id, measure), df in data.groupby(['cause_id', 'measure_id']):
            cause = causes_map[c_id]
            measure = 'yll' if measure == MEASURES['YLLs'] else 'yld'
            df = utilities.filter_data_by_restrictions(df, cause, measure, utility_data.get_age_group_ids())
            temp.append(df)
        data = pd.concat(temp, ignore_index=True)

    else:  # etiology
        data = extract.extract_data(entity, 'etiology_population_attributable_fraction', location_id)
        cause = [c for c in causes if entity in c.etiologies][0]
        data = utilities.filter_data_by_restrictions(data, cause, 'inner', utility_data.get_age_group_ids())
        if np.any(data[DRAW_COLUMNS] < 0):
            logger.warning(f"{entity.name.capitalize()} has negative values for paf. These will be replaced with 0.")
            other_cols = [c for c in data.columns if c not in DRAW_COLUMNS]
            data.set_index(other_cols, inplace=True)
            data = data.where(data[DRAW_COLUMNS] > 0, 0).reset_index()

    data = utilities.convert_affected_entity(data, 'cause_id')
    data.loc[data['measure_id'] == MEASURES['YLLs'], 'affected_measure'] = 'excess_mortality_rate'
    data.loc[data['measure_id'] == MEASURES['YLDs'], 'affected_measure'] = 'incidence_rate'
    data = (data.groupby(['affected_entity', 'affected_measure'])
            .apply(utilities.normalize, fill_value=0)
            .reset_index(drop=True))
    data = data.filter(DEMOGRAPHIC_COLUMNS + ['affected_entity', 'affected_measure'] + DRAW_COLUMNS)
    return data


def get_estimate(entity: Covariate, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'estimate', location_id)

    key_columns = ['location_id', 'year_id']
    if entity.by_age:
        key_columns.append('age_group_id')
    if entity.by_sex:
        key_columns.append('sex_id')

    data = data.filter(key_columns + COVARIATE_VALUE_COLUMNS)
    data = utilities.normalize(data)
    data = utilities.wide_to_long(data, COVARIATE_VALUE_COLUMNS, var_name='parameter')
    return data


def get_utilization_rate(entity: HealthcareEntity, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'utilization_rate', location_id)
    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_structure(entity: Population, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'structure', location_id)
    data = data.drop('run_id', 'columns').rename(columns={'population': 'value'})
    data = utilities.normalize(data)
    return data


def get_theoretical_minimum_risk_life_expectancy(entity: Population, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'theoretical_minimum_risk_life_expectancy', location_id)
    data = data.rename(columns={'age': 'age_start', 'life_expectancy': 'value'})
    data['age_end'] = data.age_start.shift(-1).fillna(125.)
    return data


def get_age_bins(entity: Population, location_id: int) -> pd.DataFrame:
    age_bins = utility_data.get_age_bins()[['age_group_name', 'age_start', 'age_end']]
    return age_bins


def get_demographic_dimensions(entity: Population, location_id: int) -> pd.DataFrame:
    demographic_dimensions = utility_data.get_demographic_dimensions(location_id)
    demographic_dimensions = utilities.normalize(demographic_dimensions)
    return demographic_dimensions
